import elastic from '@elastic/elasticsearch';
import { v4 as uuid } from 'uuid';
import { mapFilter } from 'microtil';
import { filterToAccess, authorizedByPermission, forbidden, notFound } from './backendAbstract.js';
const Client = elastic.Client;
let clientInstance;
export const client = () => clientInstance || init();
export const destroyClient = () => {
    clientInstance = undefined;
};
export const init = () => {
    const node = process.env.ELASTIC_HOST;
    const username = process.env.ELASTIC_USER_NAME;
    const password = process.env.ELASTIC_PASSWORD;
    const apiKey = process.env.ELASTIC_API_KEY;
    const apiId = process.env.ELASTIC_API_ID;
    const unauthenticated = process.env.ELASTIC_UNAUTHENTICATED;
    const setup = { node, requestTimeout: 90000 };
    if (username && password) {
        setup.auth = { username, password };
    }
    else if (apiKey) {
        setup.auth = apiId ? { apiKey: { id: apiId, api_key: apiKey } } : { apiKey };
    }
    else if (!unauthenticated) {
        console.warn('Elasticsearch api credentials are not set');
    }
    clientInstance = new Client(setup);
    return clientInstance;
};
export const info = () => client().info();
export const createIndex = (name) => client().indices.create({ index: name });
export const defaultSize = 64;
const getById = async (index, id) => {
    const { body: { _source } } = await client().get({ index, id });
    return _source;
};
const getByIdChecked = async (index, id, authentication, auth, manageFields) => filterToAccess([await getById(index, id)], authentication, auth, manageFields);
export const get = async (contract, auth, idIn, input) => {
    var _a, _b;
    const index = contract.implementation.index.toLowerCase();
    const { manageFields, authentication: authDef } = contract;
    const userIdFilter = {
        match: {
            createdBy: auth.sub || ''
        }
    };
    const id = idIn || ((_a = input) === null || _a === void 0 ? void 0 : _a.id);
    const search = (_b = input) === null || _b === void 0 ? void 0 : _b.search;
    if (Array.isArray(id)) {
        if (id.length === 0)
            return { result: [] };
        const { body: { docs } } = await client().mget({ index, body: { ids: id } });
        return { result: filterToAccess(mapFilter(docs, (x) => x._source), authDef, auth, manageFields) };
    }
    else if (id) {
        try {
            const got = await getByIdChecked(index, id, authDef, auth, manageFields);
            if (got.length === 0) {
                return forbidden(id);
            }
            return { result: got };
        }
        catch (got) {
            if (got.meta.statusCode === 404)
                return notFound({ id, input });
            throw got;
        }
    }
    else if (search) {
        const queryString = {
            query: {
                bool: {
                    must: [{ simple_query_string: { query: search } }]
                }
            }
        };
        if (!authorizedByPermission(authDef, auth))
            queryString.query.bool.must.push(userIdFilter);
        const all = await client().search({
            index,
            body: queryString,
            size: contract.implementation.maxResults || defaultSize
        });
        return { result: new Array(all.body.hits.hits).flatMap((y) => y.map((x) => x._source)) };
    }
    const searchAll = { index, size: contract.implementation.maxResults || defaultSize };
    if (!authorizedByPermission(authDef, auth)) {
        searchAll.body = {
            query: {
                bool: {
                    must: [userIdFilter]
                }
            }
        };
    }
    const all = await client().search(searchAll);
    const result = new Array(all.body.hits.hits).flatMap((y) => y.map((x) => x._source));
    // console.log('IN THE END', result)
    return { result: result };
};
export const post = async (contract, auth, id, body) => {
    const idNew = id || uuid();
    const newBody = { ...body };
    if (contract.manageFields.id === true) {
        newBody.id = idNew;
    }
    if (contract.manageFields.createdBy === true) {
        newBody.createdBy = auth.sub;
    }
    try {
        await client().create({
            id: idNew,
            index: contract.implementation.index.toLowerCase(),
            refresh: 'wait_for',
            body: newBody
        });
    }
    catch (e) {
        if (e.meta.statusCode === 409)
            return { errorType: 'conflict', data: body, status: 409, errors: [] };
        throw e;
    }
    return { result: newBody };
};
export const del = async (contract, auth, id) => {
    const index = contract.implementation.index.toLowerCase();
    if (Array.isArray(id)) {
        const data = await Promise.all(id.map(async (x) => ({ id, result: await del(contract, auth, x) })));
        const errors = data.reduce((p, c) => p.concat(Array.isArray(c.result.errors) && c.result.errors.length
            ? c.result.errors
            : []), []);
        if (errors.length) {
            return { errorType: 'forbidden', data, status: 403, errors };
        }
        return { result: {} };
    }
    try {
        const result = await getByIdChecked(index, id, contract.authentication, auth, contract.manageFields);
        if (!result || result.length === 0) {
            return forbidden(id, ['Can\'t delete, unauthorized']);
        }
        await client().delete({ index, id, refresh: 'wait_for' });
    }
    catch (error) {
        if (error.meta.statusCode === 404)
            return { errorType: 'notFound', status: 404, data: id, errors: [] };
        throw error;
    }
    return { result: {} };
};
export const patch = async (contract, auth, id, body) => {
    const index = contract.implementation.index.toLowerCase();
    try {
        const result = await getByIdChecked(index, id, contract.authentication, auth, contract.manageFields);
        if (!result || result.length === 0)
            return forbidden({ id, body });
        let newBody = body;
        if (contract.manageFields.createdBy === true) {
            newBody = { ...body };
            newBody.createdBy = result[0].createdBy;
        }
        await client().update({
            index,
            refresh: 'wait_for',
            id,
            body: { doc: newBody }
        });
    }
    catch (error) {
        if (error.meta.statusCode === 404)
            return notFound({ id, body });
        throw error;
    }
    return { result: {} };
};
export const put = async (contract, auth, id, body) => {
    const index = contract.implementation.index.toLowerCase();
    try {
        const result = await getByIdChecked(index, id, contract.authentication, auth, contract.manageFields);
        if (!result || result.length === 0)
            return forbidden({ id, body });
        let newBody = body;
        if (contract.manageFields.createdBy === true) {
            newBody = { ...body };
            newBody.createdBy = result[0].createdBy;
        }
        await client().index({
            index,
            refresh: 'wait_for',
            id,
            body: newBody
        });
        return { result: {} };
    }
    catch (error) {
        if (error.meta.statusCode === 404)
            return notFound({ id, body });
        throw error;
    }
};
export const getElasticsearchProvider = () => ({
    get,
    post,
    put,
    patch,
    delete: del
});
//# sourceMappingURL=backendElasticsearch.js.map