import { v4 as uuid } from 'uuid';
import { memoryKV } from './memoryKv.js';
import { workerKv } from './workerKv.js';
import { forbidden, notFound } from './backendAbstract.js';
const clientInstance = {};
const typeToString = (input) => typeof input === 'string' ? input : input.custom;
export const client = (key) => clientInstance[typeToString(key)] || init(typeToString(key));
export const destroyAllClients = () => {
    for (const key of Object.keys(clientInstance))
        delete clientInstance[key];
};
export const destroyClient = (key) => {
    delete clientInstance[key];
};
export const init = (key) => {
    if (key === 'worker') {
        clientInstance.worker = workerKv();
        return clientInstance.worker;
    }
    if (key === 'memory') {
        clientInstance.memory = memoryKV();
        return clientInstance.memory;
    }
    if (typeof customKv !== 'undefined') {
        if (key in customKv) {
            clientInstance[key] = customKv[key]();
            return clientInstance[key];
        }
    }
    throw new Error(`Unknown key value backend: '${key}'`);
};
const authorizedByPermission = (auth, authInput) => typeof auth === 'boolean' ||
    auth.some(x => (authInput.permissions || []).some(y => x === y));
const getUserIdFields = (fields) => Object.entries(fields).filter(x => x[1]).map(x => x[0]);
const filterToAccess = (input, auth, authInput, fields) => authorizedByPermission(auth, authInput) ? input : input.filter((x) => getUserIdFields(fields).some(y => x[y] === authInput.sub));
const keyId = (index, id) => `${index}:records:${id}`;
const getByIdChecked = async (id, auth, type, index, authDef, manageFields) => {
    const result = await client(type).get(keyId(index, id), 'json');
    return filterToAccess([result], authDef, auth, manageFields);
};
export const get = async (contract, auth, idIn, input) => {
    var _a, _b, _c;
    const id = idIn || ((_a = input) === null || _a === void 0 ? void 0 : _a.id);
    let cursor = (_b = input) === null || _b === void 0 ? void 0 : _b.cursor;
    const limit = ((_c = input) === null || _c === void 0 ? void 0 : _c.limit) || 64;
    const type = contract.implementation.backend;
    const index = contract.implementation.prefix;
    if (Array.isArray(id)) {
        if (id.length === 0)
            return { result: [] };
        const docs = (await Promise.all(id.map(x => client(type).get(keyId(index, x), 'json'))))
            .filter(x => x != null);
        return { result: filterToAccess(docs, contract.authentication, auth, contract.manageFields) };
    }
    else if (id) {
        const result = await client(type).get(keyId(index, id), 'json');
        if (!result)
            return notFound({ id, input });
        const filtered = filterToAccess([result], contract.authentication, auth, contract.manageFields);
        if (filtered.length === 0)
            return forbidden(input);
        return { result: filtered };
    }
    if (!contract.implementation.allowGetAll)
        return { errorType: 'badInput', status: 400, errors: ['Get all is disabled, id must be provided'] };
    const accessAll = authorizedByPermission(contract.authentication, auth);
    const listId = [];
    const result = await client(type)
        .list({ limit: Math.max(10, limit), cursor, prefix: `${index}:records` });
    result.keys.forEach(async (x) => {
        if (accessAll || x.metadata.createdBy === auth.sub) {
            listId.push(client(type).get(x.name, 'json'));
        }
    });
    if (listId.length >= limit)
        cursor = null;
    if (result.list_complete)
        cursor = null;
    return {
        result: (await Promise.all(listId)).filter((x) => x != null),
        cursor: result.cursor,
        more: !result.list_complete
    };
};
export const post = async (contract, auth, id, body) => {
    const type = contract.implementation.backend;
    const index = contract.implementation.prefix;
    const newId = id || uuid();
    const newBody = { ...body };
    if (contract.manageFields.id === true) {
        newBody.id = newId;
    }
    const metadata = {};
    if (contract.manageFields.createdBy === true) {
        newBody.createdBy = auth.sub;
        metadata.createdBy = auth.sub;
    }
    // Maybe skip check if it is generated?
    const got = await client(type).get(keyId(index, newId));
    if (got)
        return { errorType: 'conflict', data: body, status: 409, errors: [] };
    // TODO returned without the full id, that contains the index, or maybe always remove the index when returning?
    await client(type).put(keyId(index, newId), JSON.stringify(newBody), { metadata });
    return { result: newBody };
};
export const del = async (contract, auth, id) => {
    if (Array.isArray(id)) {
        const data = await Promise.all(id.map(async (x) => ({ id, result: await del(contract, auth, x) })));
        const errors = data.reduce((p, c) => p.concat(Array.isArray(c.result.errors) && c.result.errors.length
            ? c.result.errors
            : []), []);
        if (errors.length) {
            return { errorType: 'forbidden', data, status: 403, errors };
        }
        return { result: {} };
    }
    const type = contract.implementation.backend;
    const index = contract.implementation.prefix;
    const result = await getByIdChecked(id, auth, type, index, contract.authentication, contract.manageFields);
    if (result.length === 1 && result[0] === null) {
        return notFound({ id });
    }
    if (!result || result.length === 0)
        return forbidden(id, [`forbidden - could not delete item: ${id} `]);
    await client(type).delete(keyId(index, id));
    return { result: {} };
};
export const patch = async (contract, auth, id, body) => {
    const type = contract.implementation.backend;
    const index = contract.implementation.prefix;
    const result = await getByIdChecked(id, auth, type, index, contract.authentication, contract.manageFields);
    if (!result || result.length === 0)
        return forbidden({ id, body });
    const newBody = { ...result[0] };
    for (const [key, value] of Object.entries(body)) {
        newBody[key] = value;
    }
    if (contract.manageFields.createdBy === true) {
        newBody.createdBy = result[0].createdBy;
    }
    const key = keyId(index, id);
    const { value, metadata } = await client(type).getWithMetadata(key);
    if (value == null) {
        return notFound({ id, body });
    }
    await client(type).put(key, JSON.stringify(newBody), { metadata });
    return { result: {} };
};
export const put = async (contract, auth, id, body) => {
    const type = contract.implementation.backend;
    const index = contract.implementation.prefix;
    const result = await getByIdChecked(id, auth, type, index, contract.authentication, contract.manageFields);
    if (!result || result.length === 0)
        return forbidden({ id, body });
    const newBody = { ...body };
    if (contract.manageFields.createdBy === true) {
        newBody.createdBy = result[0].createdBy;
    }
    const key = keyId(index, id);
    const { value, metadata } = await client(type).getWithMetadata(key);
    if (value == null) {
        return notFound({ id, body });
    }
    await client(type).put(key, JSON.stringify(newBody), { metadata });
    return { result: {} };
};
export const getKvProvider = () => ({
    get,
    post,
    put,
    patch,
    delete: del
});
//# sourceMappingURL=backendKv.js.map